
from ignite.metrics import Average
import numpy as np
import numbers
import torch
from typing import Any, Callable, Optional, Union
from ignite.metrics.metric import Metric, reinit__is_reduced, sync_all_reduce


class SafeAverage(Average):
    @sync_all_reduce("accumulator", "num_examples")
    def compute(self) -> Union[Any, torch.Tensor, numbers.Number]:
        if self.num_examples < 1:
            return torch.from_numpy(np.array([np.nan], dtype=np.float32))
            # raise NotComputableError(
            #    "{} must have at least one example before" " it can be computed.".format(self.__class__.__name__)
            # )

        return self.accumulator / self.num_examples



import argparse

from aws_sagemaker_remote.training.args import sagemaker_training_args
from aws_sagemaker_remote.training.main import sagemaker_training_run, sagemaker_env_args
from aws_sagemaker_remote.processing.args import sagemaker_processing_args
from aws_sagemaker_remote.processing.main import sagemaker_processing_handle
from .config import IgniterConfig
from aws_sagemaker_remote.commands import Command
from aws_sagemaker_remote.training.main import TrainingCommand
from pytorch_igniter.trainer import train, RunSpec
from .args import train_kwargs, bool_argument
from pytorch_igniter.evaluator import evaluate


def mlflow_args(
        parser,
        mlflow_enable=True,
        mlflow_tracking_uri=None,
        mlflow_tracking_password=None,
        mlflow_tracking_username=None,
        mlflow_tracking_secret_name=None,
        mlflow_tracking_secret_region=None,
        mlflow_tracking_secret_profile=None,
        mlflow_experiment_name='default',
        mlflow_run_name=None):
    bool_argument(parser, '--mlflow-enable', default=mlflow_enable,
                  help='Enable logging to MLflow (default: {})'.format(mlflow_enable))
    parser.add_argument('--mlflow-experiment-name', default=mlflow_experiment_name,
                        help='Experiment name in MLflow (default: {})'.format(mlflow_experiment_name))
    parser.add_argument('--mlflow-run-name', default=mlflow_run_name,
                        help='Run name in MLflow (default: {})'.format(mlflow_run_name))
    parser.add_argument('--mlflow-tracking-uri', default=mlflow_tracking_uri,
                        help='URI of MLflow tracking server (default: ``{}``)'.format(mlflow_tracking_uri))
    parser.add_argument('--mlflow-tracking-username', default=mlflow_tracking_username, type=str,
                        help='Username for MLflow tracking server (default: ``{}``)'.format(mlflow_tracking_username))
    parser.add_argument('--mlflow-tracking-password', default=mlflow_tracking_password, type=str,
                        help='Password for MLflow tracking server (default: ``{}``)'.format(mlflow_tracking_password))
    parser.add_argument('--mlflow-tracking-secret-name', default=mlflow_tracking_secret_name, type=str,
                        help='Secret for accessing MLflow (default: ``{}``)'.format(mlflow_tracking_secret_name))
    parser.add_argument('--mlflow-tracking-secret-profile', default=mlflow_tracking_secret_profile, type=str,
                        help='Profile for accessing secret for accessing MLflow (default: ``{}``)'.format(mlflow_tracking_secret_profile))
    parser.add_argument('--mlflow-tracking-secret-region', default=mlflow_tracking_secret_region, type=str,
                        help='Region for accessing secret for accessing MLflow (default: ``{}``)'.format(mlflow_tracking_secret_region))


def train_args(parser, max_epochs=10, n_saved=10):
    parser.add_argument(
        '--max-epochs', type=int, default=max_epochs, metavar='N',
        help='number of epochs to train (default: {})'.format(max_epochs))
    parser.add_argument(
        '--n-saved', default=n_saved, type=int,
        help='Number of checkpoints to keep (default: ``{}``)'.format(n_saved))
    # parser.add_argument('--seed', type=int, default=1, metavar='S',
    #                        help='random seed (default: 1)')


def model_args(parser, device=None):
    parser.add_argument("--device", type=str, default=device,
                        help="device to use (default: {})".format(device))


class TrainCommand(TrainingCommand):
    def runner(self, args):
        model = self.igniter_config.make_model(args).to(args.device)
        trainer = self.igniter_config.make_trainer(args, model).to(args.device)
        to_save = {
            'user_trainer': trainer,
            'model': model
        }
        #to_save['model'] = model
        train(
            to_save=to_save,
            train_spec=trainer.spec,
            # todo: check kwargs
            **train_kwargs(args),
            parameters=vars(args),
            # ,
            inference_args=args,
            inference_spec=self.igniter_config.inference_spec,
            model=model
            # =args
        )

    def argparse_callback(self, parser):
        group = parser.add_argument_group(
            title='Model',
            description='Model arguments'
        )
        model_args(group)
        if self.igniter_config.model_args:
            self.igniter_config.model_args(group)
        group = parser.add_argument_group(
            title='Training',
            description='Training arguments'
        )
        train_args(group)
        if self.igniter_config.train_args:
            self.igniter_config.train_args(group)
        parser.add_argument(
            '--cmd', default=self.cmd, help=argparse.SUPPRESS
        )
        group = parser.add_argument_group(
            title='MLflow',
            description='MLflow arguments'
        )
        mlflow_args(group)

    def __init__(self, igniter_config: IgniterConfig, script, cmd='train', **kwargs):
        super(TrainCommand, self).__init__(
            help='Train a model',
            main=self.runner,
            script=script,
            inputs=igniter_config.train_inputs,
            argparse_callback=self.argparse_callback,
            **kwargs
        )
        self.igniter_config = igniter_config
        self.train_config = None
        self.script = script
        self.cmd = cmd


class EvalCommand(TrainingCommand):
    def runner(self, args):
        model = self.igniter_config.make_model(args).to(args.device)
        evaluator = self.igniter_config.make_evaluator(
            args, model)
        evaluate(
            eval_spec=evaluator.spec,
            to_load={
                'model': model
            },
            model_dir=args.model_dir,
            output_dir=args.output_dir
            # todo: check kwargs
            # **train_kwargs(args),
            # parameters=vars(args)
        )

    def argparse_callback(self, parser):
        group = parser.add_argument_group(
            title='Model',
            description='Model arguments'
        )
        model_args(group)
        if self.igniter_config.model_args:
            self.igniter_config.model_args(group)
        group = parser.add_argument_group(
            title='Evaluation',
            description='Evaluation arguments'
        )
        if self.igniter_config.eval_args:
            self.igniter_config.eval_args(group)
        parser.add_argument(
            '--cmd', default=self.cmd, help=argparse.SUPPRESS
        )
        group = parser.add_argument_group(
            title='MLflow',
            description='MLflow arguments'
        )
        mlflow_args(group)

    def __init__(self, igniter_config: IgniterConfig, script, cmd='eval', **kwargs):
        super(EvalCommand, self).__init__(
            help='Evaluate a model',
            script=script,
            main=self.runner,
            inputs=igniter_config.eval_inputs,
            argparse_callback=self.argparse_callback,
            **kwargs
        )
        self.igniter_config = igniter_config
        self.script = script
        self.cmd = cmd

    # def configure(self, parser: argparse.ArgumentParser):
    #    super(EvalCommand, self).configure(parser)
    # def run(self, args):
    #    pass


class TrainAndEvalCommand(TrainingCommand):
    def runner(self, args):
        model = self.igniter_config.make_model(args).to(args.device)
        trainer = self.igniter_config.make_trainer(
            args, model).to(args.device)
        evaluator = self.igniter_config.make_evaluator(
            args, model)
        to_save = {
            'user_trainer': trainer,
            'model': model
        }
        train(
            to_save=to_save,
            model=model,
            train_spec=trainer.spec,
            eval_spec=evaluator.spec,
            # todo: check kwargs
            **train_kwargs(args),
            parameters=vars(args),
            inference_args=args,
            inference_spec=self.igniter_config.inference_spec,
        )

    def argparse_callback(self, parser):
        group = parser.add_argument_group(
            title='Model',
            description='Model arguments'
        )
        model_args(group)
        if self.igniter_config.model_args:
            self.igniter_config.model_args(group)
        group = parser.add_argument_group(
            title='Training',
            description='Training arguments'
        )
        train_args(group)
        if self.igniter_config.train_args:
            self.igniter_config.train_args(group)
        group = parser.add_argument_group(
            title='Evaluation',
            description='Evaluation arguments'
        )
        if self.igniter_config.eval_args:
            self.igniter_config.eval_args(group)
        parser.add_argument(
            '--cmd', default=self.cmd, help=argparse.SUPPRESS
        )
        group = parser.add_argument_group(
            title='MLflow',
            description='MLflow arguments'
        )
        mlflow_args(group)

    def __init__(self, igniter_config: IgniterConfig, script, cmd='train-and-eval', **kwargs):
        super(TrainAndEvalCommand, self).__init__(
            help='Train and evaluate a model',
            script=script,
            main=self.runner,
            argparse_callback=self.argparse_callback,
            inputs={
                **igniter_config.train_inputs,
                **igniter_config.eval_inputs
            },
            **kwargs
        )
        self.cmd = cmd
        self.igniter_config = igniter_config
        self.script = script

    # def configure(self, parser: argparse.ArgumentParser):
    #     super(TrainAndEvalCommand, self).configure(parser)



import os
import mlflow
import contextlib
from ignite.contrib.handlers.mlflow_logger import MLflowLogger
import yaml


class NullContext(object):
    def __enter__(self):
        pass

    def __exit__(self, type, value, tb):
        pass


RUN_FNAME = 'run.yaml'


def mlflow_ctx(
    checkpoint_dir=None,
    output_dir=None,
    run_id=None,
    mlflow_enable=True,
    allow_new=True,
    experiment_name=None,
    run_name=None,
    parameters=None,
    is_sagemaker=None,
    sagemaker_job_name=None
):
    if mlflow_enable:
        # Check for a run already in progress
        active_run = mlflow.active_run()
        if active_run is not None:
            print("MLflow ID {} active".format(active_run.info.run_id))
            return NullContext()
        # Use argument
        if run_id is not None:
            print("MLflow ID {} set".format(run_id))
            return mlflow.start_run(run_id=run_id)
        # Check saved run id
        if checkpoint_dir is not None:
            run_fname = os.path.join(checkpoint_dir, RUN_FNAME)
            if os.path.exists(run_fname):
                with open(run_fname) as f:
                    run_id = yaml.load(f, Loader=yaml.SafeLoader)[
                        'info']['run_id']
                print("MLflow ID {} from run file".format(run_id))
                return mlflow.start_run(run_id=run_id)
        # New run
        if allow_new:
            print("MLflow new run")
            if experiment_name:
                try:
                    experiment = mlflow.get_experiment_by_name(
                        name=experiment_name)
                    if experiment:
                        experiment_id = experiment.experiment_id
                    else:
                        experiment_id = mlflow.create_experiment(
                            name=experiment_name)
                except Exception as e:
                    print(e)
                    experiment_id = mlflow.create_experiment(
                            name=experiment_name)
                    
                #todo: wait for experiment to be fully created. otherwise start_run fails
            else:
                experiment_id = None
            ctx = mlflow.start_run(
                run_id=run_id, experiment_id=experiment_id, run_name=run_name)
            if parameters:
                mlflow.log_params(parameters)
            if is_sagemaker and sagemaker_job_name:
                mlflow.set_tag('SageMakerJobName', sagemaker_job_name)
            return ctx
        else:
            raise ValueError("No existing MLflow run found")
    else:
        return NullContext()


def get_mlflow_logger(output_dir=None, checkpoint_dir=None, mlflow_enable=True):
    if mlflow_enable:
        mlflow_logger = MLflowLogger()
        active_run = mlflow.active_run()
        active_run = mlflow.get_run(active_run.info.run_id)
        if output_dir is not None:
            run_fname = os.path.join(output_dir, RUN_FNAME)
            with open(run_fname, 'w') as f:
                yaml.dump(active_run.to_dictionary(), f)
        if checkpoint_dir is not None and output_dir != checkpoint_dir:
            run_fname = os.path.join(checkpoint_dir, RUN_FNAME)
            with open(run_fname, 'w') as f:
                yaml.dump(active_run.to_dictionary(), f)
        return mlflow_logger
    else:
        return None
